# coding:utf-8
# Copyright (c) 2022  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .base_handler import BaseTaskflowHandler


class TaskflowHandler(BaseTaskflowHandler):
    def __init__(self):
        self._name = "taskflow_handler"

    @classmethod
    def process(cls, predictor, data, parameters):
        if data is None:
            return {}
        text = None
        if "text" in data:
            text = data["text"]
        else:
            return {}
        if "schema" in parameters:
            schema = parameters["schema"]
            predictor.set_schema(schema)
        return predictor(text)
