# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .bleu import BLEU, BLEUForDuReader
from .chunk import ChunkEvaluator
from .distinct import Distinct
from .glue import AccuracyAndF1, Mcc, MultiLabelsMetric, PearsonAndSpearman
from .mrr import MRR
from .perplexity import Perplexity
from .rouge import Rouge1, Rouge2, RougeL, RougeLForDuReader, RougeN
from .sighan import CorrectionF1, DetectionF1
from .span import SpanEvaluator
