# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .bellegroup import *
from .cail2018_small import *
from .cblue import *
from .chnsenticorp import *
from .clue import *
from .cmrc2018 import *
from .conll2002 import *
from .cote import *
from .couplet import *
from .dataset import *
from .drcd import *
from .drcd_cn import *
from .dureader_robust import *
from .glue import *
from .imdb import *
from .intokens_dataset import *
from .lcqmc import *
from .msra_ner import *
from .nlpcc13_evsam05_hit import *
from .nlpcc13_evsam05_thu import *
from .nlpcc14_sc import *
from .nlpcc_dbqa import *
from .peoples_daily_ner import *
from .poetry import *
from .ptb import *
from .seabsa16 import *
from .squad import *
from .wmt14ende import *
from .wos import *
from .xnli import *
from .xnli_cn import *
from .yahoo_answer_100k import *
