# THIS FILE IS GENERATED FROM PADDLENLP SETUP.PY
commit           = '49520692fb80898fd53ab708831ae54bb53073b7'

__all__ = ['show']

def show():
    """Get the corresponding commit id of paddlenlp.

    Returns:
        The commit-id of paddlenlp will be output.

        full_version: version of paddlenlp


    Examples:
        .. code-block:: python

            import paddlenlp

            paddlenlp.version.show()
            # commit: 1ef5b94a18773bb0b1bba1651526e5f5fc5b16fa

    """
    print("commit:", commit)

